# FILE: misc.tcl
#
# Various Tcl/Tk utility scripts for the OOMMF 2D display program
#

# Procedure to append a bindtag at the end of a window's
# curreng bindtag list
proc OMF_AppendBindTag { win newtag } {
    bindtags $win [concat [bindtags $win] $newtag]
}

# Automatically select/deselect entire text upon Entry widget FocusIn/Out
#bind Entry <FocusIn> {+%W selection from 0 ; %W selection to end }
bind Entry <FocusOut> {+%W selection clear}
# I took out these binding on 97/2/16, because I don't like the
# effect.  -mjd
# It appears that Tcl 8.0b1 does the <FocusIn> binding by default,
# but not the <FocusOut>, which I _really_ dislike, so I reinstated
# the latter on 97/9/26.   -mjd

set omfStdNumericFormat "%4.2f"

proc OMF_FormatNumber { var {fmt {}}} {
    if { [string match {} $fmt] } {
	global omfStdNumericFormat
	set fmt $omfStdNumericFormat
    }
    return [format $fmt $var]
}

# Load input filter program options
set omfMmdispFilter [list [list *.gz {gzip -dc}]]
Oc_Option Get Input filters omfMmdispFilter

# In order to get case-insensitive matching, all filenames are converted
# to lowercase before matching.  So, we insure all filterstring patterns
# are lowercase as well:
set lowered_list {}
foreach eltpair $omfMmdispFilter {
    if {[llength $eltpair]!=2} {
	Ow_FatalError "Illegal mmDisp input filter pair: $eltpair" \
		"Error setting up omfMmdispFilter"
    }
    set filterstring [string tolower [lindex $eltpair 0]]
    set filterprog   [lindex $eltpair 1]
    lappend lowered_list [list $filterstring $filterprog]
}
set omfMmdispFilter $lowered_list
unset lowered_list

# Proc to filter input files.  Returns name of temporary file
# if filtering done, null string if no filtering is done (either
# because no filter is set, or because the input file doesn't match
# the glob string.
# IMPORTANT NOTE: Is is the resposibility of the calling procedure
#   to delete the temporary file when it is no longer needed.
# ANOTHER IMPORTANT NOTE: This code ASSUMES omfMmdispFilter filterstrings
#   are lowercase, and lowercases all input filenames for matching.
proc OMF_MmdispInputFilter { infile } {
    global omfMmdispFilter 
    if {"$omfMmdispFilter"==""} {return {}}  ;# Filter not set
    set usefilter {}
    set infilelow [string tolower $infile]
    foreach eltpair $omfMmdispFilter {
	if {[llength $eltpair]!=2} {
	    Ow_FatalError "Illegal mmDisp input filter pair: $eltpair" \
		    "Error in OMF_MmdispInputFilter"
	}
	set filterstring [lindex $eltpair 0]
	set filterprog   [lindex $eltpair 1]
	foreach i $filterstring {
	    # Index through glob strings
	    if [string match $i $infilelow] {
		set usefilter $filterprog
		break
	    }
	}
	if {"$usefilter"!=""} break
    }
    if {"$usefilter"==""} {return {}}   ;# No match found

    # Otherwise, we have a match.  Make up temporary file name
    if {[catch {Oc_TempFile New t -stem mmdff}]} {
	Ow_NonFatalError "Unable to create unique temporary filename using \
		basename mmdff" "Error in OMF_MmdispInputFilter"
	return {}
    } else {
        set tempname [$t AbsoluteName]
        $t Delete
    }

    # Try to run filter command.  Note that we only want one
    # substitution pass on $infile and $tempname, lest they have
    # spaces or backslash sequences in their values.
    if [catch {eval exec $filterprog {<$infile >$tempname}} errmsg] {
	Ow_NonFatalError "Error while trying to execute \
		$filterprog: $errmsg" "Error in OMF_MmdispInputFilter"
	catch {file delete $tempname}
    }

    if [file exists $tempname] { return $tempname } ;# Apparent success
    return {}  ;# Failure
}


# Trough color for 'scale' widgets and 'scrollbar' widgets.
# By default, Windows sets these to same as default frame background.
# set omfTroughColor #b8b8b8
scale .omftemp
set temp [winfo rgb . [.omftemp cget -background]]
set tempr [expr round([lindex $temp 0]*0.85)]
set tempg [expr round([lindex $temp 1]*0.85)]
set tempb [expr round([lindex $temp 2]*0.85)]
set omfTroughColor [format "#%04X%04X%04X" $tempr $tempg $tempb]
unset temp tempr tempg tempb
destroy .omftemp

# Get a list of all descendents of a window (children, grandchildren, ...)
proc OMF_GetDescendentList { basewin } {
    set winlist {}
    set childlist [winfo children $basewin]
    foreach w $childlist {
	lappend winlist $w
	eval lappend winlist [OMF_GetDescendentList $w]
    }
    return $winlist
}

# Get a list of all current toplevel windows.  If import
# $include_hidden is true, returns complete list.  Otherwise
# returns only those toplevel windows that are visible (viewable).
proc OMF_GetTopLevelList { {include_hidden 0} } {
    set toplist "."
    foreach w [OMF_GetDescendentList .] {
	if { [string compare $w [winfo toplevel $w]] == 0 } {
	    if { $include_hidden || [winfo viewable $w] } {
		lappend toplist $w
	    }
	}
    }
    return $toplist
}

# Push/Pop stack operators
proc OMF_Push { stackname newvalue } {
    upvar $stackname stack
    lappend stack $newvalue
}
proc OMF_Pop { stackname } {
    # NOTE: Returns {} if stack is empty
    upvar $stackname stack
    if { ![info exists stack] } { return {} }
    set size [llength $stack]
    if { $size < 1 } { return {} }
    incr size -1
    set top_elt [lindex $stack $size]
    set stack [lreplace $stack $size $size]
    return $top_elt
}

# Check if $x is a valid base-ten number
proc OMF_IsNumber { x } {
    # Returns 1 iff $x is a valid number

    # Don't allow octal representations
    if { [regexp {^ *[-+]?0+ *$} "$x"] } then { return 1 } ;# This is a 0
    if { [regexp {^ *[-+]?0[^.]*$} "$x"] } then { return 0 }

    # If not octal, check for valid forms
    if { [regexp {^ *[-+]?[0-9]+\.?[0-9]* *$} "$x"] } then { return 1 }
    if { [regexp {^ *[-+]?[0-9]*\.?[0-9]+ *$} "$x"] } then { return 1 }

    # Else, illegal form.
    return 0
}

# Return a modified string by removing anything that is not
# 0-9 or '.', and keeping only first '.'.
# NOTE: This routine *may* return the string ".", which is
#       *not* a legal number for most (any?) standard
#       Tcl/Tk routines.  This makes for more amiable
#       interactive editing of a traced variable, however.
proc OMF_MakeValidPosNumber { string } {
    regsub -all -nocase {[^0-9.]} $string {} stringfix
    # Now remove all '.''s except the first.  Do this by
    # temporarily changing the first '.' to an 'x'
    if [regsub {\.} $stringfix {x} stringfix] {
        regsub -all {\.} $stringfix {} stringfix
        regsub {x} $stringfix {.} stringfix
    }
    return $stringfix
}

# A trace proc providing OMF_MakeValidPosNumber service.
proc OMF_KeepValidPosNumber { name elt op } {
    global mmd_log
    if { ![string match {} $elt] } { set name ${name}($elt) }
    upvar $name x
    if {![info exists x]} {
	# Try global level
	upvar #0 $name x
	if {![info exists x]} {
	    $mmd_log Append "Error in proc OMF_KeepValidPosNumber, \
		    run level #[info level]: \
		    Unable to locate variable $name"
	    return {}
	}
    }
    set x [OMF_MakeValidPosNumber $x]
}

# Similar to OMF_MakeValidPosNumber, but allows a leading '-'.
# NOTE: This routine *may* return the string ".", which is
#       *not* a legal number for most (any?) standard
#       Tcl/Tk routines.  This makes for more amiable
#       interactive editing of a traced variable, however.
proc OMF_MakeValidNumber { string } {
    regsub -all -nocase {[^-0-9.]} $string {} stringfix
    # Now remove all '.''s except the first.  Do this by
    # temporarily changing the first '.' to an 'x'
    if [regsub {\.} $stringfix {x} stringfix] {
        regsub -all {\.} $stringfix {} stringfix
        regsub {x} $stringfix {.} stringfix
    }
    # Next strip out any "-"'s except leading one, if any.
    regexp -- {^(-?)(.*)} $stringfix junk prefix number
    regsub -all -- {-} $number {} number
    return $prefix$number
}

# A trace proc providing OMF_MakeValidNumber service.
proc OMF_KeepValidNumber { name elt op } {
    global mmd_log
    if { ![string match {} $elt] } { set name ${name}($elt) }
    upvar $name x
    if {![info exists x]} {
	# Try global level
	upvar #0 $name x
	if {![info exists x]} {
	    $mmd_log Append "Error in proc OMF_KeepValidNumber, \
		    run level #[info level]: \
		    Unable to locate variable $name"
	    return {}
	}
    }
    set x [OMF_MakeValidNumber $x]
}

# Checks if x & y would differ if converted the the omfStdNumericFormat.
# Used <FocusOut> bindings to determine if the shadowed variable needs
# updating.  Returns 1 iff x!=y after conversion.
proc OMF_StdFormatCompare { x y } {
    global omfStdNumericFormat
    set x [format $omfStdNumericFormat $x]
    set y [format $omfStdNumericFormat $y]
    if {$x==$y} { return 0 }
    return 1
}

# Converts width, height, xcenter & ycenter across rotations
proc OMF_RotateDimensions { angle widthname heightname \
        xcentername ycentername } {
    upvar $widthname   width
    upvar $heightname  height
    upvar $xcentername xcenter
    upvar $ycentername ycenter
    set irotstep [expr round($angle/90.)%4]
    set curwidth $width;     set curheight $height
    set curxcenter $xcenter; set curycenter $ycenter
    switch -exact $irotstep {
        1 { # 90 degrees
            set width   $curheight
            set height  $curwidth
            set xcenter $curycenter
            set ycenter [expr 1. - $curxcenter]
        }
        2 { # 180 degrees
            set xcenter [expr 1. - $curxcenter]
            set ycenter [expr 1. - $curycenter]
        }
        3 { # 270 degrees
            set width   $curheight
            set height  $curwidth
            set xcenter [expr 1. - $curycenter]
            set ycenter $curxcenter
        }
    }  ;# No changes if $irotstep==0

    # Note: There is some "grittiness" in the 'x/yview moveto'
    #   command, probably due to the scrollbar stepsize resolution
    #   (cf. the 'delta' scrollbar command).  As a result, sometimes
    #   the above calculations cause the slider bar to be off by 1
    #   step from where it should be.
}

####################################################################
# Find dimensions of root window
proc GetRootDimensions { widthname heightname } {
    upvar $widthname width
    upvar $heightname height
    set width 0;  set height 0
    regexp {^([^x]*)x([^-+]*).*$} [wm geometry .] match width height
}

####################################################################
# Proc to wait for a window to become visible.
# Timeouts after "timeout" milliseconds.  Default is 1000.
# Returns 1 on timeout, 0 otherwise.  Returns immediately
# if [winfo viewable $win] is true.
#
proc WaitVisible { win {timeout 1000} {timecheck 100}} {
    if { ![winfo exists $win] } { return 2 }
    if { [winfo viewable $win] } { return 0 }

    update idletasks
    if { [winfo viewable $win] } { return 0 }

    global WaitVisibleFlag WaitVisibleCheckFlag
    set WaitVisibleFlag "Wait"

    set id_timeout [after $timeout { set WaitVisibleFlag "Timeout" }]
    while { [string match $WaitVisibleFlag "Wait"] } {
        set WaitVisibleCheckFlag 0
        after $timecheck { set WaitVisibleCheckFlag 1 }
	vwait WaitVisibleCheckFlag
	if { [winfo viewable $win] } { 
	    set WaitVisibleFlag "Viewable"
	}
    }
    after cancel $id_timeout

    if { [string match $WaitVisibleFlag "Timeout"] } { return 1 }
    return 0
}

#####
# Info message dialog box (non-modal)
#  The return value is the window designation
#  Set import bitmap to "none" to get no bitmap
proc OMF_InfoDialog { message {title {}} {bitmap {}} {width {}} \
	{justification "left"} } {
    # First make up a unique window name
    set id 0
    set basename .omfInfoDialogWin
    while { [winfo exists "$basename$id"] } {
	if { $id > 200 } {
	    Ow_FatalError "Too many (>$id) open InfoDialog windows"
	}
	incr id
    }
    set window $basename$id

    if { {} == $title } { set title "Informational Message" }

    toplevel $window
    wm group $window .
    wm title $window "$title"
    Ow_PositionChild $window

    global omfDialogWidth
    if { {} == "$width" } { set width $omfDialogWidth }  ;# Default width

    frame $window.top
    frame $window.bottom
    # Insert beam to insure minimum width
    frame $window.top.beam -height 0 -width $width
    pack $window.top.beam -side top
    if { {} == $bitmap } { set bitmap info }
    if { [string compare "none" $bitmap] != 0 } then {
        label $window.top.bitmap -bitmap $bitmap
        pack $window.top.bitmap -side left -anchor w -padx 2m
    }

    message $window.top.msg -width $width -text $message \
	    -justify $justification
    ## UNFORTUNATELY, the message widget doesn't support "selections". :-[

    pack $window.top.msg -side left -fill both -expand 1

    set current_focus [focus]
    if { [string match {} $current_focus] } {
	set current_focus "."
    } else {
	set current_focus [winfo toplevel [focus]]
    }
    button $window.bottom.btn -text "OK" \
	    -command "OMF_SetFocus $current_focus $window; destroy $window"

    # Note: It is important to release focus _before_
    #  destroying the window, as otherwise the focus can
    #  get misplaced.  -mjd 2/97
    # Note 2: Forcing the focus against the window manager can
    #  make for some strangeness.  With the above code, when the
    #  message box is closed, the focus jumps to the window that
    #  had the focus when the message box was created (or ".";
    #  OMF_SetFocus falls back to "." if the import window doesn't
    #  exist).  I may decide to disable this in the future. -mjd 9/97
    wm protocol $window WM_DELETE_WINDOW "$window.bottom.btn invoke"
    pack $window.bottom.btn -side top -expand 1
    pack $window.top $window.bottom -side top -fill both -expand 1
    Ow_SetIcon $window

    OMF_InfoDialogHelper $window
    # WARNING WILL ROBINSON: This routine trys to control the
    # focus in a sensible manner.  To do this, it waits for
    # $window.bottom.btn to become viewable, which may require
    # event loop processing.  This can cause REENTRANCY PROBLEMS
    # IN EVENT BINDINGS.  You have been warned.

    return $window
}

proc OMF_InfoDialogHelper { window } {
    if {![winfo exists $window.bottom.btn]} {
	return
    }
    WaitVisible $window.bottom.btn

    set current_focus [focus]
    if { [string match {} $current_focus] } {
	set current_focus "."
    } else {
	set current_focus [winfo toplevel [focus]]
    }
    $window.bottom.btn configure \
	    -command "OMF_SetFocus $current_focus $window; destroy $window"
    focus $window.bottom.btn
}



# Routine to direct mouse & keyboard focus to a specified
# top-level window, provided it is not iconified (or withdrawn).
# If called by a window "destroy" routine, it is important to
# call this _before_ the window is destroyed, as otherwise the
# focus may be "misplaced" (i.e., lost).  Also if you specify
# the "other top-level window", then tlw will be raised with
# respect to that window.  (Note: Linux/AXP + XFree is painfully
# slow on 'raise $tlw', but 'raise $tlw $otlw' is quick.
proc OMF_SetFocus { tlw {otlw ""} } {
    if { ![winfo exists $tlw] } { set tlw "." }
    set top [winfo toplevel $tlw]
    if {[string compare [wm state $top] normal]==0} {
        if {[string compare $otlw ""]!=0 && [winfo exists $otlw]} {
            raise $top $otlw
        }
        focus [focus -lastfor $tlw]  ;# Return focus to last window
                                      # in $tlw that had focus
    }
}

# Convert .'s in $window to _'s, and extract everything up to
# the second underscore as a (per-call) ID that can be used to
# form global variable names.  This proc assumes that $window
# begins with a leading '.'.  This routine also extracts the
# base (up to second .) and extension part of $window
# NOTE: If "$id" is sent in place of $window, then this routine
# will return "$id" unaltered (i.e. "OMF_GetID $w" ==
# "OMF_GetID [OMF_GetID $w]")
proc OMF_GetID { window {idname ""} {basename ""} {extname ""} } {
    if { "$idname" != "" }   { upvar $idname id }
    if { "$basename" != "" } { upvar $basename base }
    if { "$extname" != "" }  { upvar $extname ext }
    set base "";
    set id "$window" ;# If regexp doesn't match, we get $window back
    regexp {^ *(\.[^.]*)(.*)$} "$window" full base ext
    regsub -all {\.} "$base" _ id
    return $id
}

# OMF_EntryFollow copies $name to ${name}Entry
proc OMF_EntryFollow { name elt op } {
    # Assumes $elt == ""
    upvar #0 $name base
    global mmd_log
    if ![info exists base] then {
        $mmd_log Append \
          "Error in OMF_EntryFollow; global variable ->$name<- does not exist"
        return
    }
    upvar #0 ${name}Entry follower
    if ![info exists follower] then {
        $mmd_log Append -nonewline "Error in OMF_EntryFollow;"
        $mmd_log Append " global variable ->${name}Entry<- does not exist"
        return
    }
    set follower $base
}

# OMF_ScaleFollow copies $name to ${name}Scale
proc OMF_ScaleFollow { name elt op } {
    # Assumes $elt == ""
    upvar #0 $name base
    upvar #0 ${name}Scale follower
    global mmd_log
    if ![info exists base] then {
        $mmd_log Append \
          "Error in OMF_ScaleFollow; global variable ->$name<- does not exist"
        return
    }
    if ![info exists follower] then {
        $mmd_log Append -nonewline "Error in OMF_ScaleFollow;"
        $mmd_log Append " global variable ->${name}Scale<- does not exist"
        return
    }
    set follower $base
}

# OMF_ScaleEntryFollow copies $name to ' regsub {Scale$} "$name"  "Entry" '
proc OMF_ScaleEntryFollow { name elt op } {
    # Assumes $elt == ""
    upvar #0 $name base
    global mmd_log
    # Make sure this routine isn't be called from OMF_ScaleFollow,
    # because in that case $name is "follower"
    if { ![info exist base] } return
    regsub {Scale$} "$name"  "Entry" followername
    upvar #0 $followername follower
    if ![info exists follower] then {
        $mmd_log Append -nonewline "Error in OMF_ScaleEntryFollow;"
        $mmd_log Append " global variable ->$followername<- does not exist"
        return
    }
    global omfStdNumericFormat
    set follower [format $omfStdNumericFormat $base]
}

# Produce an entry window collection:
#          $framename, $framename.label, & $framename.value
# You also get $framename.scrx if '$scrollbar' is true.
# If you set "numeric" to non-zero, then the displayed value is held
# in a shadow variable called ${varname}Entry, at the global level.
# It is copied the "varname" upon a <Key-Return> or a <FocusOut> event,
# and the value from "varname" is copied back into ${varname}Entry on
# an <Key-Escape> event, or anytime "varname" changes.
# Note: Variables tied to entry widgets are at the global level.  Also,
#  bindings run at global level, so either way "varname" must be at
#  the global level.
# If "numeric" is >0, then the value is kept a valid positive number
# (via OMF_KeepValidPosNumber), or if "numeric" <0 then it is kept at
# any valid number (via OMF_KeepValidNumber).
proc OMF_MakeEntryWindow { framename label labelwidth varname varwidth \
        scrollbar {relief "raised"} {numeric "0"}} {
    global $varname
    frame $framename -relief $relief -bd 2
    label $framename.label -text "$label"
    if {[string compare $labelwidth ""]!=0} {
        $framename.label configure -width $labelwidth -anchor e
    }
    entry $framename.value -relief sunken -bd 2 \
            -width $varwidth -bg [$framename.label cget -bg]
    # The default color scheme under Tcl 8.0 + Windows makes
    # labels a light gray, and entry backgrounds bright white.
    # *I* think it looks pretty hideous...  -mjd 97/10/3

    OMF_AppendBindTag $framename.value omfCursorOverride
    pack $framename.label -side left -anchor e
    if { $scrollbar } then {
        $framename.value configure \
                -xscrollcommand "$framename.scrx set"
        global omfTroughColor
        scrollbar $framename.scrx -bd 2 -troughcolor $omfTroughColor \
                -orient horizontal -command "$framename.value xview"
        pack $framename.scrx -side bottom -fill x
    }
    if { $numeric != 0 } {
	global ${varname}Entry
	$framename.value configure -textvariable ${varname}Entry \
		-justify right -font [Oc_Font Get bold]
	if { $numeric > 0 } {
	    trace variable ${varname}Entry w OMF_KeepValidPosNumber
	} else {
	    trace variable ${varname}Entry w OMF_KeepValidNumber
	}
	bind $framename.value <Key-Return> \
		"set $varname \[set ${varname}Entry\]"
	bind $framename.value <FocusOut> \
		"set $varname \[set ${varname}Entry\]"
	bind $framename.value <Key-Escape> \
		"set ${varname}Entry \[set $varname\]"
	trace variable $varname w OMF_EntryFollow
    } else {
	$framename.value configure -textvariable $varname
    }
    pack $framename.value -fill x -expand 1 -side left
}

# Produce a entry+scale window collection
proc OMF_MakeEntryScaleWindow {framename label labelwidth varname varwidth \
        scalewidth scalemin scalemax scalestep procname} {
    # NOTE: varname *must* be a global variable name
    # Make widgets
    frame $framename -borderwidth 2 -relief ridge
    frame $framename.col1
    frame $framename.col2 -bd 2 -relief sunken
    label $framename.label -text $label -anchor e
    if {[string compare $labelwidth {}]!=0} {
        $framename.label configure -width $labelwidth
    }
    entry $framename.value -textvariable ${varname}Entry \
        -width $varwidth -bd 0 -justify right \
	 -bg [$framename.label cget -bg]
    # The default color scheme under Tcl 8.0 + Windows makes
    # labels a light gray, and entry backgrounds bright white.
    # *I* think it looks pretty hideous...  -mjd 97/10/3

    OMF_AppendBindTag $framename.value omfCursorOverride

    $framename.value configure -font [Oc_Font Get bold]

    global omfTroughColor
    scale $framename.scale -variable ${varname}Scale \
        -orient horizontal -troughcolor $omfTroughColor -showvalue 0 \
        -length $scalewidth -from $scalemin -to $scalemax \
        -resolution  $scalestep -bigincrement 0 -bd 0
    # Pack widgets
    pack $framename.label -side left  -in $framename.col1
    pack $framename.value -side left  -in $framename.col2
    pack $framename.scale -side right -in $framename.col2 -fill x -expand 1
    pack $framename.col1 -side left
    pack $framename.col2 -side right -fill x -expand 1
    # Bind widgets
    bind $framename.value <Key-Return> \
            "OMF_EntryScaleBinding $procname $varname"
    bind $framename.value <FocusOut> \
            "OMF_EntryScaleBinding $procname $varname 1"
    bind $framename.value <Key-Escape> "OMF_EntryScaleEscapeBinding $varname"
    bind $framename.scale <ButtonRelease> \
            "OMF_EntryScaleButtonReleaseBinding %W %x %y $procname  $varname"
    bind $framename.scale <Leave> \
            "tkCancelRepeat ; OMF_EntryScaleEscapeBinding $varname"
    # Note: tkCancelRepeat is an apparently undocumented call that turns
    # off the auto-stepping feature initiated by a <Button-1> event in the
    # scale trough.  If this breaks in a new release of Tk, then another
    # approach might be to try using "%W config -state disabled/normal" calls.
    bind $framename.scale <Motion> \
            "OMF_ScaleMotionBinding %W %x %y"
    bind $framename.scale <Key-Return> \
            "OMF_ScaleScaleBinding %W $procname $varname 0"
    bind $framename.scale <FocusOut> \
            "OMF_ScaleScaleBinding %W $procname $varname 1"
    bind $framename.scale <Key-Escape> "OMF_EntryScaleEscapeBinding $varname"
    bind $framename.scale <Button-2> "set ${varname}Scale \[%W get %x %y\]"
    # Set traces
    global $varname ${varname}Entry ${varname}Scale
    ### Tie *Entry variables to *Scale variables
    trace variable ${varname}Scale w OMF_ScaleEntryFollow
    ### Tie *Entry & *Scale variables to base variables
    trace variable $varname w OMF_EntryFollow
    trace variable $varname w OMF_ScaleFollow
    ### Make sure *Entry variable gets written in 'Standard Numeric Format'
    if {![info exists $varname]} { set $varname 0.0 }
    set $varname [OMF_FormatNumber [set $varname]]
    set ${varname}Entry [set $varname]
}

proc OMF_EntryScaleBinding { procname varname {stepcheck 0}} {
    global $varname ${varname}Entry
    # Make sure we got valid input from user
    if { [OMF_IsNumber [set ${varname}Entry]] } {
        if {!$stepcheck || [OMF_StdFormatCompare \
                [set $varname] [set ${varname}Entry]]} {
            $procname [set ${varname}Entry]
        }
    } else {
        set ${varname}Entry [set $varname]
    }
    set ${varname}Entry [OMF_FormatNumber [set $varname]]
}

proc OMF_ScaleScaleBinding { widget procname varname {stepcheck 0}} {
    upvar #0 $varname var
    upvar #0 ${varname}Scale varscale
    if {!$stepcheck || 2*abs($var-$varscale) > [$widget cget -resolution]} {
            $procname $varscale
    }
}

proc OMF_EntryScaleEscapeBinding { varname } {
    global $varname ${varname}Entry ${varname}Scale
    global omfStdNumericFormat
    set ${varname}Scale [set $varname]
    set ${varname}Entry [format $omfStdNumericFormat [set $varname]]
    # Note: Order is important here.  There is a trace on ${varname}Scale
    #       that resets ${varname}Entry, and some Tcl/Tk interpreters
    #       will force ${varname}Scale to the nearest scale resolution
    #       step, which then gets copied to ${varname}Entry by the tracing.
    #       So we must set ${varname}Entry *after* ${varname}Scale.
}

proc OMF_EntryScaleButtonReleaseBinding { w x y procname varname } {
    # Updates $varname if button release was over slider or trough,
    # otherwise calls OMF_EntryScaleEscapeBinding
    global $varname ${varname}Entry ${varname}Scale
    if {[string compare [$w identify $x $y] {}]!=0} {
        $procname [set ${varname}Scale]
    } else {
        OMF_EntryScaleEscapeBinding $varname
    }
}

proc OMF_ScaleMotionBinding { w x y } {
    if {[string compare [$w identify $x $y] {}]==0} {return -code break}
}

# Produce a radio box window collection
# Each elt of "args" is a label/value list.  Any args elt
# with only 1 sub-element is automatically defined a value
# equivalent to its label, but all lowercase.
# Ex:  OMF_MakeRadioList frame "Sample:" "" sample {} \
#            A {B 2} "C Button" {"Last button" 4}
#  gets label/values of: A/a B/2 "C Button"/"c button" "Last button"/4
proc OMF_MakeRadioList {framename rows center label labelwidth \
        varname tracename args} {
    # Note: If you want centering inside the frame, do
    #    pack $framename.label -side left -anchor e -expand 1
    # and on the last button do
    #    pack $framename.r$j -side left -anchor w -expand 1
    frame $framename -borderwidth 2 -relief ridge
    label $framename.label -text "$label" -anchor e
    if {[string compare $labelwidth {}]!=0} {
        $framename.label configure -width $labelwidth
    }
    if {$rows<2} {
        pack $framename.label -side left
    } else {
        pack $framename.label -side top
    }
    if {$center} {
        pack configure $framename.label -anchor e -expand 1
    }
    set j 0
    foreach i $args {
        if {[llength $i]>2} {
            Ow_FatalError "Too long list element: $i" \
                    "OMF_MakeRadioList Error"
        }
        set text ""
        if {[llength $i]>0} {
            set text [lindex $i 0]
        }
        if {[llength $i]>1} {
            set value [lindex $i 1]
        } else {
            set value [string tolower $text]
        }
        radiobutton $framename.r$j -text $text -variable $varname \
                -value $value
        if {$rows<2} {
            pack $framename.r$j -side left -anchor w
        } else {
            pack $framename.r$j -side left -expand 1
        }
        incr j
    }
    if {$center} {
        pack configure $framename.r[expr $j-1] -expand 1
    }
    if {[string compare $tracename {}]!=0} {
        global $varname; trace variable $varname w $tracename
    }
}

# Converts a length specification of the form 4.5i (4.5 inches) to an
# equivalent length in specified units, and returns the converted length
# *without* a suffix.  Valid units/suffixes are: c (cm), i (inch), m (mm),
# and p for printer's points (=1/72 inch).  If no suffix is found then
# p is assumed.  This is consistent with the interpretation used by Tk's
# canvas widget's 'postscript' command.  The default output unit is
# centimeters.
proc OMF_ConvertUnits { length {newunits "c"}} {
    set length [string trim $length]
    if ![regexp {^(-?[0-9.]+)(.?)} $length match number dimension] {
        Ow_FatalError "Invalid length: $length" "OMF_ConvertToCm Error"
    }
    # Determine multiplier to convert to cm
    switch -exact -- [string tolower $dimension] {
        c { set multa 1. }
        i { set multa 2.54 }
        m { set multa 0.1 }
        {} -
        p { set multa [expr 2.54/72.] }
        default {
            Ow_FatalError "Invalid dimension ($dimension) on length: \
                    $length" "OMF_ConvertUnits Error"
        }
    }
    # Determine multiplier to convert from cm to "newunits"
    switch -exact -- [string tolower $newunits] {
        c { set multb 1. }
        i { set multb 1./2.54 }
        m { set multb 10. }
        {} -
        p { set multb [expr 72./2.54] }
        default {
            Ow_FatalError "Invalid output unit request: $newunits" \
                    "OMF_ConvertUnits Error"
        }
    }
    return [expr $number*$multa*$multb]
}

# Print a canvas
#   The input is an array including the following fields:
#    canvas     -   name of canvas widget
#    inname     -   name of file being displayed
#    outname    -   name of output file or pipe
#    orient     -   portrait | landscape
#    size       -   output width *or* height (not both), in the format
#                     '-pagewidth 3.i'  (3 inches width), or
#                     '-pageheight 5.c' (5 cm high), for example
#    papertype  -   letter | legal | A4 | A3
#    clip       -    1 | 0  (1=>print only displayed region of canvas)
#    xmargin    -    left/right margin, with units.  E.g.: 1.i
#    ymargin    -    top/bottom margin,  as above
#    position   -    One of: c nw n ne e se s sw w
proc OMF_PrintCanvas { parr } {
    upvar $parr popt
    set errcode 0

    # File overwrite check
    set popt(outname) [string trim $popt(outname)]
    if { "|" != [string index $popt(outname) 0] \
	    && [file exists $popt(outname)] } {
	# File is not a pipe, and already exists.
	# Check with user before overwriting.
	global ProgName
	set response [Ow_Dialog 1 "$ProgName: Print Warning" \
		warning "Overwrite existing file $popt(outname)?" \
		{} "1" "Yes" "No"]
	if { $response != 0 } {
	    # First (0'th) button (which is "Yes") _not_ selected
	    set errcode 1
	    return $errcode
	}
    }

    # Calculate positioning (in cm)
    switch -exact -- [string tolower $popt(papertype)] {
        # Set paper dimensions
        letter { set xmax 8.5i;  set ymax 11i }
        legal  { set xmax 8.5i;  set ymax 14i }
        a4     { set xmax 21.c;  set ymax 29.7c }
        a3     { set xmax 29.7c; set ymax 42.c  }
        default { Ow_FatalError "Illegal papertype: $popt(papertype)" \
                "OMF_PrintCanvas Error" }
    }
    set xmax [OMF_ConvertUnits $xmax]
    set ymax [OMF_ConvertUnits $ymax]
    set xmargin [OMF_ConvertUnits $popt(xmargin)]
    set ymargin [OMF_ConvertUnits $popt(ymargin)]
    if {[string compare "landscape" [string tolower $popt(orient)]]==0} {
        # Rotation notes: If $rotate==1, then the 'page' coordinate system
        # origin is moved to the upper lefthand corner of the page, with
        # +x running down the page and +y running to the right.  If you
        # adopt this reference frame, then you can ignore the rotation
        # and just pretend you are working from the start on a landscape'd
        # page.  The man page and online documentation
        # http://www.sunlabs.com/tcl/development/man/TkCmd/canvas.n.html#M55
        # says "in rotated output the x-axis runs along the long dimension
        # of the page (``landscape'' orientation)."  This is true for
        # the 'canvas' coordinate system, but not true for the 'page'
        # coordinate system (adjusted with the -pagex & -pagey options).
        set rotate 1
        set temp $xmargin ; set xmargin $ymargin; set ymargin $temp
    } else { set rotate 0 }
    # The following page offsets are used in conjuction
    # with the '$canvas postscript -anchor' option
    if {"$popt(position)"==""} { set $popt(position) "c" } ;# Safety
    if {[string first "n" $popt(position)]>=0} {
        # Position at top of page, minus margin
        {!$rotate} {
            set yoff [expr $ymax-$ymargin]
        } else {
            set xoff $xmargin
        }
    } elseif {[string first "s" $popt(position)]>=0} {
        # Position along page bottom margin
        if {!$rotate} {
            set yoff $ymargin
        } else {
            set xoff [expr $xmax-$xmargin]
        }
    } else {
        # Top/bottom center
        if {!$rotate} {
            set yoff [expr $ymax/2.]
        } else {
            set xoff [expr $xmax/2.]
        }
    }
    if {[string first "e" $popt(position)]>=0} {
        # Position page right, minus margin
        if {!$rotate} {
            set xoff [expr $xmax-$xmargin]
        } else {
            set yoff [expr $ymax-$ymargin]
        }
    } elseif {[string first "w" $popt(position)]>=0} {
        # Position page left, plus margin
        if {!$rotate} {
            set xoff $xmargin
        } else {
            set yoff $ymargin
        }
    } else {
        # Left/right center
        if {!$rotate} {
            set xoff [expr $xmax/2.]
        } else {
            set yoff [expr $ymax/2.]
        }
    }

    # Setup crop options
    if {$popt(crop)} {
        set cropoptions {}
    } else {
        # Printing whole canvas
        set sr [.canvas_frame.canvas cget -scrollregion]
        set width [expr [lindex $sr 2] - [lindex $sr 0]]
        set height [expr [lindex $sr 3] - [lindex $sr 1]]
        set cropoptions "-x 0 -y 0 -width $width -height $height"
    }


    # Print
    set output {}
    set errmsg {}
    if { [catch {
	set output [open $popt(outname) w]
	puts -nonewline $output \
		[eval $popt(canvas) postscript -colormode color \
		$popt(size) -pageanchor $popt(position) \
		-pagex ${xoff}c -pagey ${yoff}c -rotate $rotate \
		$cropoptions ]
    } errmsg] } {
	global ProgName
	Ow_NonFatalError "Error printing to $popt(outname): $errmsg" \
		"$ProgName: Print Error"
	set errcode 2
    }
    catch {
	if { ![string match {} $output] } { close $output }
    }

    return $errcode
}

